//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2003, 2004 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//
#include "common.hh"

#include "inst/program_builder.hh"
#include "vm/cpu.hh"

using vm::Cpu;

namespace inst
{

  // Arithmetics instructions ------------------------------

  void
  ProgramBuilder::add_abs(Register *dest, Register *src)
  {
    // FIXME: this move should be in the delay slot
    program_->text_section ().
      add_inst(new Addu(*dest, *src, Register(Register::general, Cpu::zero)));
    program_->text_section ().add_inst(new Bgez(*src, new IntExp(4)));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll
               (Register(Register::general, Cpu::zero),
                Register(Register::general, Cpu::zero),
                new IntExp(0)));
    program_->text_section ().
      add_inst(new Sub(*dest, Register(Register::general, Cpu::zero), *src));
  }

  void
  ProgramBuilder::add_mul(Register *dest, Register *src1, Exp *src2)
  {
    // Warning: is not implemented in as for mips, check why.
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Mul(*dest, *src1, Register(Register::general, Cpu::at)));
  }

  void
  ProgramBuilder::add_div(Register *dest, Register *src1, Register *src2)
  {
    // FIXME: break on divide by zero, check overflow, see mips-as output
    program_->text_section ().add_inst(new Div(*src1, *src2));
    program_->text_section ().add_inst(new Mflo(*dest));
  }
  void
  ProgramBuilder::add_div(Register *dest, Register *src1, Exp *src2)
  {
    // FIXME: program_solver must warn and/or break when exp is zero
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Div(*src1, Register(Register::general, Cpu::at)));
    program_->text_section ().add_inst(new Mflo(*dest));
  }
  void
  ProgramBuilder::add_divu(Register *dest, Register *src1, Register *src2)
  {
    // FIXME: break on divide by zero, check overflow, see mips-as output
    program_->text_section ().add_inst(new Divu(*src1, *src2));
    program_->text_section ().add_inst(new Mflo(*dest));
  }
  void
  ProgramBuilder::add_divu(Register *dest, Register *src1, Exp *src2)
  {
    // FIXME: program_solver must warn and/or break when exp is zero
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Divu(*src1, Register(Register::general, Cpu::at)));
    program_->text_section ().add_inst(new Mflo(*dest));
  }

  void
  ProgramBuilder::add_rem(Register *dest, Register *src1, Register *src2)
  {
    // FIXME: break on divide by zero, check overflow, see mips-as output
    program_->text_section ().add_inst(new Div(*src1, *src2));
    program_->text_section ().add_inst(new Mfhi(*dest));
  }
  void
  ProgramBuilder::add_remu(Register *dest, Register *src1, Register *src2)
  {
    // FIXME: break on divide by zero, see mips-as output
    program_->text_section ().add_inst(new Divu(*src1, *src2));
    program_->text_section ().add_inst(new Mfhi(*dest));
  }


  // Bitwise instructions -------------------------------------

  void
  ProgramBuilder::add_rol(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().
      add_inst(new Subu(Register(Register::general, Cpu::at),
                        Register(Register::general, Cpu::zero), *src2));
    program_->text_section ().
      add_inst(new Srlv(Register(Register::general, Cpu::at), *src1,
                        Register(Register::general, Cpu::at)));
    program_->text_section ().add_inst(new Sllv(*dest, *src1, *src2));
    program_->text_section ().
      add_inst(new Or(*dest, *dest, Register(Register::general, Cpu::at)));
  }

  void
  ProgramBuilder::add_rol(Register *dest, Register *src1, Exp *src2)
  {
    CountedExp counted_exp(src2);

    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::at), *src1,
                       counted_exp));
    program_->text_section ().
      add_inst(new Srl(*dest, *src1,
                       new OpExp(new IntExp(32), OpExp::sub, counted_exp)));
    program_->text_section ().
      add_inst(new Or(*dest, *dest, Register(Register::general, Cpu::at)));
  }
  void
  ProgramBuilder::add_ror(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().
      add_inst(new Subu(Register(Register::general, Cpu::at),
                        Register(Register::general, Cpu::zero), *src2));
    program_->text_section ().
      add_inst(new Sllv(Register(Register::general, Cpu::at),
                        *src1, Register(Register::general, Cpu::at)));
    program_->text_section ().add_inst(new Srlv(*dest, *src1, *src2));
    program_->text_section ().
      add_inst(new Or(*dest, *dest, Register(Register::general, Cpu::at)));
  }
  void
  ProgramBuilder::add_ror(Register *dest, Register *src1, Exp *src2)
  {
    CountedExp counted_exp(src2);

    program_->text_section ().
      add_inst(new Srl(Register(Register::general, Cpu::at), *src1,
                       counted_exp));
    program_->text_section ().
      add_inst(new Sll(*dest, *src1,
                       new OpExp(new IntExp(32), OpExp::sub, counted_exp)));
    program_->text_section ().
      add_inst(new Or(*dest, *dest, Register(Register::general, Cpu::at)));
  }

  void
  ProgramBuilder::add_nor(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().add_inst(new Ori(*dest, *src1, src2));
    program_->text_section ().
      add_inst(new Nor(*dest, *dest, Register(Register::general, Cpu::zero)));
  }

  // Test instructions -------------------------------------

  // Equal
  void
  ProgramBuilder::add_seq(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().add_inst(new Xor(*dest, *src1, *src2));
    program_->text_section ().add_inst(new Sltiu(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_seq(Register *dest, Register *src, Exp *exp)
  {
    program_->text_section ().add_inst(new Xori(*dest, *src, exp));
    program_->text_section ().add_inst(new Sltiu(*dest, *dest, new IntExp(1)));
  }

  void
  ProgramBuilder::add_sne(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().add_inst(new Xor(*dest, *src1, *src2));
    program_->text_section ().
      add_inst(new Sltu(*dest,
                        Register(Register::general, Cpu::zero), *dest));
  }
  void
  ProgramBuilder::add_sne(Register *dest, Register *src, Exp *exp)
  {
    program_->text_section ().add_inst(new Xori(*dest, *src, exp));
    program_->text_section ().
      add_inst(new Sltu(*dest, Register(Register::general, Cpu::zero), *dest));
  }

  // Greater
  void
  ProgramBuilder::add_sge(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().add_inst(new Slt(*dest, *src1, *src2));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_sge(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().add_inst(new Slti(*dest, *src1, src2));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_sgeu(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().add_inst(new Sltu(*dest, *src1, *src2));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_sgeu(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().add_inst(new Sltiu(*dest, *src1, src2));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }

  void
  ProgramBuilder::add_sgt(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Slt(*dest, Register(Register::general, Cpu::at), *src1));
  }
  void
  ProgramBuilder::add_sgtu(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Sltu(*dest, Register(Register::general, Cpu::at), *src1));
  }

  // Lower
  void
  ProgramBuilder::add_sle(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().add_inst(new Slt(*dest, *src2, *src1));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_sle(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Slt(*dest, Register(Register::general, Cpu::at), *src1));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_sleu(Register *dest, Register *src1, Register *src2)
  {
    program_->text_section ().add_inst(new Sltu(*dest, *src2, *src1));
    program_->text_section ().add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }
  void
  ProgramBuilder::add_sleu(Register *dest, Register *src1, Exp *src2)
  {
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Sltu(*dest, Register(Register::general, Cpu::at), *src1));
    program_->text_section ().
      add_inst(new Xori(*dest, *dest, new IntExp(1)));
  }


  // Branch instructions -----------------------------------

  // Equal
  void
  ProgramBuilder::add_bne(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Bne(*src1, Register(Register::general, Cpu::at), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_beq(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Addiu(Register(Register::general, Cpu::at),
                         Register(Register::general, Cpu::zero), src2));
    program_->text_section ().
      add_inst(new Beq(*src1, Register(Register::general, Cpu::at), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }

  // Greater
  void
  ProgramBuilder::add_bge(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slt(Register(Register::general, Cpu::at), *src1, *src2));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bge(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slti(Register(Register::general, Cpu::at), *src1, src2));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bgeu(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltu(Register(Register::general, Cpu::at), *src1, *src2));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bgeu(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltiu(Register(Register::general, Cpu::at), *src1, src2));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }

  void
  ProgramBuilder::add_bgt(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slt(Register(Register::general, Cpu::at), *src2, *src1));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bgt(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slti(Register(Register::general, Cpu::at), *src1,
                        new OpExp(src2, OpExp::add, new IntExp(1))));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bgtu(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltu(Register(Register::general, Cpu::at), *src2, *src1));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bgtu(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltiu(Register(Register::general, Cpu::at), *src1,
                         new OpExp(src2, OpExp::add, new IntExp(1))));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }

  // Lower
  void
  ProgramBuilder::add_ble(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slt(Register(Register::general, Cpu::at), *src2, *src1));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_ble(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slti(Register(Register::general, Cpu::at), *src1,
                        new OpExp(src2, OpExp::add, new IntExp(1))));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bleu(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltu(Register(Register::general, Cpu::at), *src2, *src1));
    program_->text_section ().
      add_inst(new Beq(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bleu(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltiu(Register(Register::general, Cpu::at), *src1,
                         new OpExp(src2, OpExp::add, new IntExp(1))));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }

  void
  ProgramBuilder::add_blt(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slt(Register(Register::general, Cpu::at), *src1, *src2));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_blt(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Slti(Register(Register::general, Cpu::at), *src1, src2));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bltu(Register *src1, Register *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltu(Register(Register::general, Cpu::at), *src1, *src2));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }
  void
  ProgramBuilder::add_bltu(Register *src1, Exp *src2, Exp *dest)
  {
    program_->text_section ().
      add_inst(new Sltiu(Register(Register::general, Cpu::at), *src1, src2));
    program_->text_section ().
      add_inst(new Bne(Register(Register::general, Cpu::at),
                       Register(Register::general, Cpu::zero), dest));
    // This nop is in the delay slot
    program_->text_section ().
      add_inst(new Sll(Register(Register::general, Cpu::zero),
                       Register(Register::general, Cpu::zero),
                       new IntExp(0)));
  }

  // Load --------------------------------------------------

  // Load address
  // FIXME: alias?
  void
  ProgramBuilder::add_la (Register *dest, Exp *exp, Register *base)
  {
    program_->text_section ().add_inst(new Addiu(*dest, *base, exp));
  }
  void
  ProgramBuilder::add_la (Register *reg, Exp *exp)
  {
    program_->text_section ().
      add_inst(new Addiu(*reg, Register(Register::general, Cpu::zero), exp));
  }

} // namespace inst
