//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2003, 2004, 2006 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Nolimips.  If not, see <http://www.gnu.org/licenses/>.
//
#include <iomanip>

#include "vm/cpu.hh"
#include "vm/cp0.hh"

#include "inst/all.hh"
#include "inst/int_exp.hh"

namespace vm
{

  // --------------------------------------------------------------------------
  // Constructor and destructor
  // --------------------------------------------------------------------------
  Cpu::Cpu(Mmu &mmu, Cp0 &cp0):
    mmu_(mmu), cp0_(cp0),
    check_callee_save_p_(false),
    trace_p_(false),
    bubble_(new inst::Sll(inst::Register(inst::Register::general, Cpu::zero),
                          inst::Register(inst::Register::general, Cpu::zero),
                          new inst::IntExp(0)))
  {
    reset();
  }

  Cpu::~Cpu()
  {
    delete bubble_;
  }

  // --------------------------------------------------------------------------
  // Internals
  // --------------------------------------------------------------------------
  void
  Cpu::reset()
  {
    // Initialize general purpose registers to 0.
    for (unsigned i = 0; i < 32; ++i)
      GPR_[i] = 0;
    // Initialize stack pointers to bottom of stack.
    GPR_[sp] = Memory::stack_bottom;
    GPR_[fp] = Memory::stack_bottom;
    // Initialize special registers to 0.
    lo_ = 0;
    hi_ = 0;
    pc_ = 0;

    // Initialize pipeline with nops
    for (unsigned i = 0; i < 6; ++i)
      pipeline_[i] = bubble_;

    counters_[0] = 0;
    counters_[1] = 0;
    counters_[2] = 0;
  }

  // --------------------------------------------------------------------------
  // Print operators
  // --------------------------------------------------------------------------
  void
  Cpu::write(std::ostream &os) const
  {
    std::ios_base::fmtflags flags = os.flags (std::ios_base::hex);

    os << "Register Map: \tpc = " << std::setfill('0') << std::setw(8) << pc_ << std::endl
         << "zero ($0) = " << std::setfill('0') << std::setw(8) << GPR_[zero]
         << "  t0 ($8) = " << std::setfill('0') << std::setw(8) << GPR_[t0]
         << "  s0($16) = " << std::setfill('0') << std::setw(8) << GPR_[s0]
         << "  t8($24) = " << std::setfill('0') << std::setw(8) << GPR_[t8] << std::endl
         << "  at ($1) = " << std::setfill('0') << std::setw(8) << GPR_[at]
         << "  t1 ($9) = " << std::setfill('0') << std::setw(8) << GPR_[t1]
         << "  s1($17) = " << std::setfill('0') << std::setw(8) << GPR_[s1]
         << "  t9($25) = " << std::setfill('0') << std::setw(8) << GPR_[t9] << std::endl
         << "  v0 ($2) = " << std::setfill('0') << std::setw(8) << GPR_[v0]
         << "  t2($10) = " << std::setfill('0') << std::setw(8) << GPR_[t2]
         << "  s2($18) = " << std::setfill('0') << std::setw(8) << GPR_[s2]
         << "  k0($26) = " << std::setfill('0') << std::setw(8) << GPR_[k0] << std::endl
         << "  v1 ($3) = " << std::setfill('0') << std::setw(8) << GPR_[v1]
         << "  t3($11) = " << std::setfill('0') << std::setw(8) << GPR_[t3]
         << "  s3($19) = " << std::setfill('0') << std::setw(8) << GPR_[s3]
         << "  k1($27) = " << std::setfill('0') << std::setw(8) << GPR_[k1] << std::endl
         << "  a0 ($4) = " << std::setfill('0') << std::setw(8) << GPR_[a0]
         << "  t4($12) = " << std::setfill('0') << std::setw(8) << GPR_[t4]
         << "  s4($20) = " << std::setfill('0') << std::setw(8) << GPR_[s4]
         << "  gp($28) = " << std::setfill('0') << std::setw(8) << GPR_[gp] << std::endl
         << "  a1 ($5) = " << std::setfill('0') << std::setw(8) << GPR_[a1]
         << "  t5($13) = " << std::setfill('0') << std::setw(8) << GPR_[t5]
         << "  s5($21) = " << std::setfill('0') << std::setw(8) << GPR_[s5]
         << "  sp($29) = " << std::setfill('0') << std::setw(8) << GPR_[sp] << std::endl
         << "  a2 ($6) = " << std::setfill('0') << std::setw(8) << GPR_[a2]
         << "  t6($14) = " << std::setfill('0') << std::setw(8) << GPR_[t6]
         << "  s6($22) = " << std::setfill('0') << std::setw(8) << GPR_[s6]
         << "  fp($30) = " << std::setfill('0') << std::setw(8) << GPR_[fp] << std::endl
         << "  a3 ($7) = " << std::setfill('0') << std::setw(8) << GPR_[a3]
         << "  t7($15) = " << std::setfill('0') << std::setw(8) << GPR_[t7]
         << "  s7($23) = " << std::setfill('0') << std::setw(8) << GPR_[s7]
         << "  ra($31) = " << std::setfill('0') << std::setw(8) << GPR_[ra] << std::endl;
    os.flags(flags);
  }

  // --------------------------------------------------------------------------
  // Arithmetic instructions
  // --------------------------------------------------------------------------

  void
  Cpu::visit(const inst::Add& add)
  {
    register_type       a = get_register(add.get_src1 ());
    register_type       b = get_register(add.get_src2 ());

    register_type       c = a + b;
    set_register(add.get_dest (), c);

    // FIXME: might be accelerated by testing only the sign bit.
    if ((a < 0 && b < 0 && c > 0) ||
        (a > 0 && b > 0 && c < 0))
      cp0_.raise_overflow();
  }
  void
  Cpu::visit(const inst::Addi& addi)
  {
    register_type       a = get_register(addi.get_src ());
    int         b = addi.get_imm ();

    register_type       c = a + b;
    set_register(addi.get_dest (), c);

    // FIXME: might be accelerated by testing only the sign bit.
    if ((a < 0 && b < 0 && c > 0) ||
        (a > 0 && b > 0 && c < 0))
      cp0_.raise_overflow();
  }
  void
  Cpu::visit(const inst::Addu& addu)
  {
    set_register(addu.get_dest (),
                 get_register(addu.get_src1 ()) +
                 get_register(addu.get_src2 ()));
  }
  void
  Cpu::visit(const inst::Addiu& addiu)
  {
    set_register(addiu.get_dest (),
                 get_register(addiu.get_src ()) + addiu.get_imm ());
  }

  void
  Cpu::visit(const inst::Sub& sub)
  {
    register_type       a = get_register(sub.get_src1 ());
    register_type       b = get_register(sub.get_src2 ());

    register_type       c = a - b;
    set_register(sub.get_dest (), c);

    if ((a < b && c > 0) || (a > b && c < 0))
      cp0_.raise_overflow();
  }
  void
  Cpu::visit(const inst::Subu& subu)
  {
    set_register(subu.get_dest (),
                 get_register(subu.get_src1 ()) -
                 get_register(subu.get_src2 ()));
  }

  void
  Cpu::visit(const inst::Sll& sll)
  {
    uregister_type      a = get_register(sll.get_src ());
    unsigned    i = sll.get_imm ();
    register_type       c = a << i;
    set_register(sll.get_dest (), c);

    // FIXME: Check overflow !
  }
  void
  Cpu::visit(const inst::Sllv& slv)
  {
    uregister_type      a = get_register(slv.get_src1 ());
    uregister_type      b = get_register(slv.get_src2 ());

    register_type       c = a << b;
    set_register(slv.get_dest (), c);

    // FIXME: Check overflow !
  }

  void
  Cpu::visit(const inst::Sra& sra)
  {
    register_type       a = get_register(sra.get_src ());
    unsigned    b = sra.get_imm ();
    register_type       c = a >> b;
    set_register(sra.get_dest (), c);

    // FIXME: Check overflow !
  }
  void
  Cpu::visit(const inst::Srav& srav)
  {
    register_type       a = get_register(srav.get_src1 ());
    register_type       b = get_register(srav.get_src2 ());

    register_type       c = a >> b;
    set_register(srav.get_dest (), c);

    // FIXME: Check overflow !
  }

  void
  Cpu::visit(const inst::Srl& srl)
  {
    uregister_type      a = get_register(srl.get_src ());
    unsigned    i = srl.get_imm ();
    uregister_type      c = a >> i;
    set_register(srl.get_dest (), c);

    // FIXME: Check overflow !
  }
  void
  Cpu::visit(const inst::Srlv& srlv)
  {
    uregister_type      a = get_register(srlv.get_src1 ());
    uregister_type      b = get_register(srlv.get_src2 ());
    uregister_type      c = a >> b;
    set_register(srlv.get_dest (), c);

    // FIXME: Check overflow !
  }


  void
  Cpu::visit(const inst::Mul& mul)
  {
    register_type       a = get_register(mul.get_src1 ());
    register_type       b = get_register(mul.get_src2 ());

    register_type       c = a * b;
    set_register(mul.get_dest (), c);

    // FIXME: Check overflow !
  }

  void
  Cpu::visit(const inst::Div& div)
  {
    register_type       a = get_register(div.get_src1 ());
    register_type       b = get_register(div.get_src2 ());

    register_type       c = a / b;
    set_lo(c);

    c = a % b;
    set_hi(c);

    // FIXME: Check overflow !
  }
  void
  Cpu::visit(const inst::Divu& divu)
  {
    uregister_type      a = get_register(divu.get_src1 ());
    uregister_type      b = get_register(divu.get_src2 ());

    set_lo(a / b);
    set_hi(a % b);
  }


  // --------------------------------------------------------------------------
  // Binary instructions
  // --------------------------------------------------------------------------

  void
  Cpu::visit(const inst::And& _and)
  {
    register_type       a = get_register(_and.get_src1 ());
    register_type       b = get_register(_and.get_src2 ());

    register_type       c = a & b;
    set_register(_and.get_dest (), c);
  }
  void
  Cpu::visit(const inst::Andi& andi)
  {
    register_type       a = get_register(andi.get_src ());
    int         b = andi.get_imm ();

    register_type       c = a & b;
    set_register(andi.get_dest (), c);
  }

  void
  Cpu::visit(const inst::Or& _or)
  {
    register_type       a = get_register(_or.get_src1 ());
    register_type       b = get_register(_or.get_src2 ());

    register_type       c = a | b;
    set_register(_or.get_dest (), c);
  }
  void
  Cpu::visit(const inst::Ori& ori)
  {
    register_type       a = get_register(ori.get_src ());
    int         b = ori.get_imm ();

    register_type       c = a | b;
    set_register(ori.get_dest (), c);
  }

  void
  Cpu::visit(const inst::Nor& nor)
  {
    register_type       a = get_register(nor.get_src1 ());
    register_type       b = get_register(nor.get_src2 ());

    register_type       c = a | b;
    set_register(nor.get_dest (), ~c);
  }

  void
  Cpu::visit(const inst::Xor& _xor)
  {
    register_type       a = get_register(_xor.get_src1 ());
    register_type       b = get_register(_xor.get_src2 ());

    register_type       c = a ^ b;
    set_register(_xor.get_dest (), c);
  }
  void
  Cpu::visit(const inst::Xori& xori)
  {
    register_type       a = get_register(xori.get_src ());
    int         b = xori.get_imm ();

    register_type       c = a ^ b;
    set_register(xori.get_dest (), c);
  }


  // --------------------------------------------------------------------------
  // Move instructions
  // --------------------------------------------------------------------------

  void
  Cpu::visit(const inst::Lui& lui)
  {
    set_register(lui.get_dest (), lui.get_imm () << 16);
  }

  // Store
  void
  Cpu::visit(const inst::Sb& sb)
  {
    register_type       addr = get_register(sb.get_base ()) + sb.get_offset ();
    mmu_.data_store_byte(addr, get_register(sb.get_src ()));
  }

  void
  Cpu::visit(const inst::Sw& sw)
  {
    register_type       addr = get_register(sw.get_base ()) + sw.get_offset ();
    mmu_.data_store_word(addr, get_register(sw.get_src ()));
  }

  // Load
  void
  Cpu::visit(const inst::Lb& lb)
  {
    register_type       addr = get_register(lb.get_base ()) + lb.get_offset ();
    set_register(lb.get_dest (), mmu_.data_load_byte(addr));
  }

  void
  Cpu::visit(const inst::Lbu& lbu)
  {
    register_type       addr = get_register(lbu.get_base ()) + lbu.get_offset ();
    unsigned    b = mmu_.data_load_byte(addr);
    set_register(lbu.get_dest (), b % 256);
  }

  void
  Cpu::visit(const inst::Lw& lw)
  {
    register_type       addr = get_register(lw.get_base ()) + lw.get_offset ();
    set_register(lw.get_dest (), mmu_.data_load_word(addr));
  }

  // --------------------------------------------------------------------------
  // Test instructions
  // --------------------------------------------------------------------------

  void
  Cpu::visit(const inst::Slt& slt)
  {
    if (get_register(slt.get_src1 ()) < get_register(slt.get_src2 ()))
      set_register(slt.get_dest(), 1);
    else
      set_register(slt.get_dest(), 0);
  }
  void
  Cpu::visit(const inst::Slti& slti)
  {
    if (get_register(slti.get_src ()) < slti.get_imm ())
      set_register(slti.get_dest(), 1);
    else
      set_register(slti.get_dest(), 0);
  }
  void
  Cpu::visit(const inst::Sltu& sltu)
  {
    if ((uregister_type) get_register(sltu.get_src1 ()) <
        (uregister_type) get_register(sltu.get_src2 ()))
      set_register(sltu.get_dest(), 1);
    else
      set_register(sltu.get_dest(), 0);
  }
  void
  Cpu::visit(const inst::Sltiu& sltiu)
  {
    if ((uregister_type) get_register(sltiu.get_src ()) <
        (uregister_type) sltiu.get_imm ())
      set_register(sltiu.get_dest(), 1);
    else
      set_register(sltiu.get_dest(), 0);
  }


  // --------------------------------------------------------------------------
  // Branch and jump instructions
  // --------------------------------------------------------------------------

  // Unconditional
  void
  Cpu::visit(const inst::Jmp& jmp)
  {
    // Jump
    set_pc(get_pc() + jmp.get_label() - 8);

    // Bubble in decode stage
    pipeline_[d_stage] = bubble_;
    // Speculatively fetched instruction
    pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
  }
  void
  Cpu::visit(const inst::Jr& jr)
  {
    // Assume it is a ret
    ret();

    // Jump
    set_pc(get_register(jr.get_dest()) - 8);

    // Bubble in decode stage
    pipeline_[d_stage] = bubble_;
    // Speculatively fetched instruction
    pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
  }

  void
  Cpu::visit(const inst::Jal& jal)
  {
    // Save return address
    set_register(Cpu::ra, get_pc());
    // Jump
    set_pc(get_register(Cpu::ra) + jal.get_label() - 8);

    // Bubble in decode stage
    pipeline_[d_stage] = bubble_;
    // Speculatively fetched instruction
    pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);

    // Assume it is a call
    call();
  }
  void
  Cpu::visit(const inst::Jalr& jalr)
  {
    // Save return address
    set_register(Cpu::ra, get_pc());
    // Jump
    set_pc(get_register(jalr.get_dest()) - 8);

    // Bubble in decode stage
    pipeline_[d_stage] = bubble_;
    // Speculatively fetched instruction
    pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);

    // Assume it is a call
    call();
  }


  // Equality
  void
  Cpu::visit(const inst::Beq& beq)
  {
    if (get_register(beq.get_src1 ()) == get_register(beq.get_src2 ()))
      {
        // Jump
        set_pc(get_pc() + beq.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }
  void
  Cpu::visit(const inst::Bne& bne)
  {
    if (get_register(bne.get_src1 ()) != get_register(bne.get_src2 ()))
      {
        // Jump
        set_pc(get_pc() + bne.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }

  // Greater
  void
  Cpu::visit(const inst::Bgez& bgez)
  {
    if (get_register(bgez.get_src ()) >= 0)
      {
        // Jump
        set_pc(get_pc() + bgez.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }
  void
  Cpu::visit(const inst::Bgezal& bgezal)
  {
    if (get_register(bgezal.get_src ()) >= 0)
      {
        // Assume it is a call
        call();

        // Save return address
        set_register(Cpu::ra, get_pc());
        // Jump
        set_pc(get_pc() + bgezal.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }
  void
  Cpu::visit(const inst::Bgtz& bgtz)
  {
    if (get_register(bgtz.get_src ()) > 0)
      {
        // Jump
        set_pc(get_pc() + bgtz.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }

  // Lower
  void
  Cpu::visit(const inst::Blez& blez)
  {
    if (get_register(blez.get_src ()) <= 0)
      {
        // Jump
        set_pc(get_pc() + blez.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }
  void
  Cpu::visit(const inst::Bltz& bltz)
  {
    if (get_register(bltz.get_src ()) < 0)
      {
        // Jump
        set_pc(get_pc() + bltz.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }
  void
  Cpu::visit(const inst::Bltzal& bltzal)
  {
    if (get_register(bltzal.get_src ()) < 0)
      {
        // Assume it is a call
        call();

        // Save return address
        set_register(Cpu::ra, get_pc());
        // Jump
        set_pc(get_pc() + bltzal.get_label() - 8);

        // Bubble in decode stage
        pipeline_[d_stage] = bubble_;
        // Speculatively fetched instruction
        pipeline_[i_stage] = & mmu_.inst_load(pc_ / 4);
      }
  }


  // --------------------------------------------------------------------------
  // Data movement instructions
  // --------------------------------------------------------------------------

  void
  Cpu::visit(const inst::Mfhi& mfhi)
  {
    set_register(mfhi.get_dest(), get_hi());
  }
  void
  Cpu::visit(const inst::Mflo& mflo)
  {
    set_register(mflo.get_dest(), get_lo());
  }
  void
  Cpu::visit(const inst::Mthi& mthi)
  {
    set_hi(get_register(mthi.get_src()));
  }
  void
  Cpu::visit(const inst::Mtlo& mtlo)
  {
    set_lo(get_register(mtlo.get_src()));
  }
  void
  Cpu::visit(const inst::Mfc0& mfc0)
  {
    set_register(mfc0.get_dest(), cp0_.get_register(mfc0.get_src()));
  }
  void
  Cpu::visit(const inst::Mtc0& mtc0)
  {
    cp0_.set_register(mtc0.get_dest(), get_register(mtc0.get_src()));
  }


  // --------------------------------------------------------------------------
  // Exception and trap instructions
  // --------------------------------------------------------------------------

  void
  Cpu::visit(const inst::Syscall&)
  {
    cp0_.raise_syscall();
  }

} // namespace vm
